#include "stdafx.h"
#include "Render3D.h"

SCALAR32 Scalar32Sin(SCALAR32 x)
{
	SCALAR32 s;
	SCALAR32 t;
	LONG i;
	t = x;
	s = t;
	i = 3;
	while(t > 0)
	{
		t = SCALAR32_MUL(t, SCALAR32_MUL(x, x)) / ((i - 1) * i);
		if(i & 2)
			s -= t;
		else
			s += t;
		i += 2;
	}
	return s;
}

SCALAR32 Scalar32Cos(SCALAR32 x)
{
	SCALAR32 s;
	SCALAR32 t;
	LONG i;
	t = SCALAR32_UNIT;
	s = t;
	i = 2;
	while(t > 0)
	{
		t = SCALAR32_MUL(t, SCALAR32_MUL(x, x)) / ((i - 1) * i);
		if(i & 2)
			s -= t;
		else
			s += t;
		i += 2;
	}
	return s;
}

void DivideVertexByW(VERTEX_XYZWUV* pVertex)
{
	if(pVertex->w != 0)
	{
		pVertex->x = SCALAR32_DIV(pVertex->x, pVertex->w);
		pVertex->y = SCALAR32_DIV(pVertex->y, pVertex->w);
		pVertex->z = SCALAR32_DIV(pVertex->z, pVertex->w);
	}
	else
	{
		pVertex->x = LONG_MIN;
		pVertex->y = LONG_MIN;
		pVertex->z = LONG_MIN;
	}
}

VERTEX_XYZWUV MultiplyMatrixVertex(MATRIX_XYZW* pMatrix, VERTEX_XYZWUV* pVertex)
{
	VERTEX_XYZWUV Temp;
	Temp.x = SCALAR32_MUL(pVertex->x, pMatrix->m[0][0]) + SCALAR32_MUL(pVertex->y, pMatrix->m[0][1]) + SCALAR32_MUL(pVertex->z, pMatrix->m[0][2]) + SCALAR32_MUL(pVertex->w, pMatrix->m[0][3]);
	Temp.y = SCALAR32_MUL(pVertex->x, pMatrix->m[1][0]) + SCALAR32_MUL(pVertex->y, pMatrix->m[1][1]) + SCALAR32_MUL(pVertex->z, pMatrix->m[1][2]) + SCALAR32_MUL(pVertex->w, pMatrix->m[1][3]);
	Temp.z = SCALAR32_MUL(pVertex->x, pMatrix->m[2][0]) + SCALAR32_MUL(pVertex->y, pMatrix->m[2][1]) + SCALAR32_MUL(pVertex->z, pMatrix->m[2][2]) + SCALAR32_MUL(pVertex->w, pMatrix->m[2][3]);
	Temp.w = SCALAR32_MUL(pVertex->x, pMatrix->m[3][0]) + SCALAR32_MUL(pVertex->y, pMatrix->m[3][1]) + SCALAR32_MUL(pVertex->z, pMatrix->m[3][2]) + SCALAR32_MUL(pVertex->w, pMatrix->m[3][3]);
	Temp.u = pVertex->u;
	Temp.v = pVertex->v;
	return Temp;
}

MATRIX_XYZW GetZeroMatrix()
{
	MATRIX_XYZW Temp = {0};
	return Temp;
}

MATRIX_XYZW GetIdentityMatrix()
{
	MATRIX_XYZW Temp = {{{SCALAR32_UNIT, 0, 0, 0}, {0, SCALAR32_UNIT, 0, 0}, {0, 0, SCALAR32_UNIT, 0}, {0, 0, 0, SCALAR32_UNIT}}};
	return Temp;
}

MATRIX_XYZW MultiplyMatrix(MATRIX_XYZW* pMatrix1, MATRIX_XYZW* pMatrix2)
{
	LONG i;
	LONG j;
	MATRIX_XYZW Temp;
	for(i = 0; i < 4; i++)
	{
		for(j = 0; j < 4; j++)
			Temp.m[i][j] = SCALAR32_MUL(pMatrix1->m[i][0], pMatrix2->m[0][j]) + SCALAR32_MUL(pMatrix1->m[i][1], pMatrix2->m[1][j]) + SCALAR32_MUL(pMatrix1->m[i][2], pMatrix2->m[2][j]) + SCALAR32_MUL(pMatrix1->m[i][3], pMatrix2->m[3][j]);
	}
	return Temp;
}

MATRIX_XYZW GetMoveMatrix(SCALAR32 x, SCALAR32 y, SCALAR32 z)
{
	MATRIX_XYZW Temp;
	Temp = GetIdentityMatrix();
	Temp.m[0][3] = x;
	Temp.m[1][3] = y;
	Temp.m[2][3] = z;
	return Temp;
}

MATRIX_XYZW GetRotateXMatrix(SCALAR32 Theta)
{
	MATRIX_XYZW Temp;
	Temp = GetIdentityMatrix();
	Temp.m[1][1] = Scalar32Cos(Theta);
	Temp.m[1][2] = Scalar32Sin(Theta) * -1;
	Temp.m[2][1] = Scalar32Sin(Theta);
	Temp.m[2][2] = Scalar32Cos(Theta);
	return Temp;
}

MATRIX_XYZW GetRotateYMatrix(SCALAR32 Theta)
{
	MATRIX_XYZW Temp;
	Temp = GetIdentityMatrix();
	Temp.m[0][0] = Scalar32Cos(Theta);
	Temp.m[0][2] = Scalar32Sin(Theta);
	Temp.m[2][0] = Scalar32Sin(Theta) * -1;
	Temp.m[2][2] = Scalar32Cos(Theta);
	return Temp;
}

MATRIX_XYZW GetRotateZMatrix(SCALAR32 Theta)
{
	MATRIX_XYZW Temp;
	Temp = GetIdentityMatrix();
	Temp.m[0][0] = Scalar32Cos(Theta);
	Temp.m[0][1] = Scalar32Sin(Theta) * -1;
	Temp.m[1][0] = Scalar32Sin(Theta);
	Temp.m[1][1] = Scalar32Cos(Theta);
	return Temp;
}

MATRIX_XYZW GetViewAngleMatrix(SCALAR32 Angle, SCALAR16 Width, SCALAR16 Height, SCALAR32 Near, SCALAR32 Far)
{
	MATRIX_XYZW Temp;
	MATRIX_XYZW Move;
	Temp = GetZeroMatrix();
	Angle = SCALAR32_DIV(Angle * max(Width, Height) * -1, SCALAR32_MAXZ16);
	Temp.m[0][0] = SCALAR32_DIV(SCALAR32_UNIT, Angle);
	Temp.m[1][1] = SCALAR32_DIV(SCALAR32_UNIT, Angle);
	Temp.m[2][2] = SCALAR32_MUL(SCALAR32_DIV(Far, Far - Near), SCALAR32_DIV(SCALAR32_MAXZ16, Far - Near));
	Temp.m[2][3] = SCALAR32_MUL(SCALAR32_DIV(SCALAR32_MUL(Near, Far), Near - Far), SCALAR32_DIV(SCALAR32_MAXZ16, Far - Near));
	Temp.m[3][2] = SCALAR32_UNIT * 2 / max(Width, Height);
	Move = GetMoveMatrix(LONG_TO_SCALAR32(Width) / 2, LONG_TO_SCALAR32(Height) / 2, 0);
	return MultiplyMatrix(&Move, &Temp);
}

BOOL CreateBitmap16Any(BITMAP16ANY* pBitmap, LONG Width, LONG Height)
{
	LONG i;
	i = 1;
	while(i < 65536)
	{
		if(Width > i && Width < (i << 1))
			Width = i << 1;
		if(Height > i && Height < (i << 1))
			Height = i << 1;
		i = i << 1;
	}
	if(Width <= 0 || Height <= 0)
		return FALSE;
	pBitmap->pBitmap = new PIXEL16[Width * Height];
	if(!pBitmap->pBitmap)
		return FALSE;
	pBitmap->Width = (SCALAR16)Width;
	pBitmap->Height = (SCALAR16)Height;
	pBitmap->MaskX = (SCALAR16)Width - 1;
	pBitmap->MaskY = (SCALAR16)Height - 1;
	pBitmap->ShiftX = 0;
	pBitmap->ShiftY = 0;
	while(Width > 1)
	{
		Width = Width >> 1;
		pBitmap->ShiftY++;
	}
	return TRUE;
}

void DestroyBitmap16Any(BITMAP16ANY* pBitmap)
{
	if(pBitmap->pBitmap)
		delete []pBitmap->pBitmap;
	pBitmap->pBitmap = NULL;
}

void DrawVertexXY(BITMAP565* pOut, VERTEX_XY* p1, VERTEX_XY* p2, VERTEX_XY* p3, PIXEL565 Color)
{
	VERTEX_XY DeltaLeft;
	VERTEX_XY DeltaRight;
	SCALAR16 x;
	SCALAR16 y;
	VERTEX_XY Left;
	VERTEX_XY Right;
	VERTEX_XY Cur;
	SCALAR16 xl;
	SCALAR16 xr;
	SCALAR16 yt;
	SCALAR16 yb;
	PIXEL565* pLine;
	BOOL bMiddleLeft;
	if(p1->y != p2->y && p1->y != p3->y)
	{
		DeltaLeft.x = SCALAR32_DIV(p2->x - p1->x, p2->y - p1->y);
		DeltaRight.x = SCALAR32_DIV(p3->x - p1->x, p3->y - p1->y);
		if(DeltaLeft.x > DeltaRight.x)
		{
			Cur = DeltaLeft;
			DeltaLeft = DeltaRight;
			DeltaRight = Cur;
			bMiddleLeft = FALSE;
		}
		else
			bMiddleLeft = TRUE;
		Left = *p1;
		Right = *p1;
		if(SCALAR32_TO_LONG(p2->y) < 0)
		{
			Left.x -= DeltaLeft.x * SCALAR32_TO_LONG(p2->y);
			Right.x -= DeltaRight.x * SCALAR32_TO_LONG(p2->y);
		}
		yt = max((SCALAR16)SCALAR32_TO_LONG(p1->y), 0);
		yb = min((SCALAR16)SCALAR32_TO_LONG(p2->y), pOut->Height);
		for(y = yt; y < yb; y++)
		{
			if(Left.x != Right.x)
			{
				xl = max((SCALAR16)SCALAR32_TO_LONG(Left.x), 0);
				xr = min((SCALAR16)SCALAR32_TO_LONG(Right.x), pOut->Width);
				pLine = pOut->pBitmap + (y << pOut->ShiftY) + xl;
				for(x = xl; x < xr; x++)
				{
					*pLine = Color;
					pLine++;
				}
			}
			Left.x += DeltaLeft.x;
			Right.x += DeltaRight.x;
		}
	}
	else
	{
		if(p1->x <= p2->x)
		{
			DeltaLeft.x = SCALAR32_DIV(p3->x - p1->x, p3->y - p1->y);
			DeltaRight.x = SCALAR32_DIV(p3->x - p2->x, p3->y - p2->y);
			Left = *p1;
			Right = *p2;
			bMiddleLeft = FALSE;
		}
		else
		{
			DeltaLeft.x = SCALAR32_DIV(p3->x - p2->x, p3->y - p2->y);
			DeltaRight.x = SCALAR32_DIV(p3->x - p1->x, p3->y - p1->y);
			Left = *p2;
			Right = *p1;
			bMiddleLeft = TRUE;
		}
	}
	if(p1->y != p3->y && p2->y != p3->y)
	{
		if(bMiddleLeft)
		{
			DeltaLeft.x = SCALAR32_DIV(p3->x - p2->x, p3->y - p2->y);
			Left = *p2;
		}
		else
		{
			DeltaRight.x = SCALAR32_DIV(p3->x - p2->x, p3->y - p2->y);
			Right = *p2;
		}
		if(SCALAR32_TO_LONG(p2->y) < 0)
		{
			Left.x -= DeltaLeft.x * SCALAR32_TO_LONG(p2->y);
			Right.x -= DeltaRight.x * SCALAR32_TO_LONG(p2->y);
		}
		yt = max((SCALAR16)SCALAR32_TO_LONG(p2->y), 0);
		yb = min((SCALAR16)SCALAR32_TO_LONG(p3->y), pOut->Height);
		for(y = yt; y < yb; y++)
		{
			if(Left.x != Right.x)
			{
				xl = max((SCALAR16)SCALAR32_TO_LONG(Left.x), 0);
				xr = min((SCALAR16)SCALAR32_TO_LONG(Right.x), pOut->Width);
				pLine = pOut->pBitmap + (y << pOut->ShiftY) + xl;
				for(x = xl; x < xr; x++)
				{
					*pLine = Color;
					pLine++;
				}
			}
			Left.x += DeltaLeft.x;
			Right.x += DeltaRight.x;
		}
	}
}

void DrawVertexXYUV(BITMAP565* pOut, BITMAP565* pTexture, VERTEX_XYUV* p1, VERTEX_XYUV* p2, VERTEX_XYUV* p3, PIXEL565* pColorKey)
{
	VERTEX_XYUV DeltaLeft;
	VERTEX_XYUV DeltaRight;
	SCALAR16 x;
	SCALAR16 y;
	VERTEX_XYUV Left;
	VERTEX_XYUV Right;
	VERTEX_XYUV Cur;
	SCALAR16 xl;
	SCALAR16 xr;
	SCALAR16 yt;
	SCALAR16 yb;
	PIXEL565* pLine;
	PIXEL565 Color;
	BOOL bMiddleLeft;
	if(p1->y != p2->y && p1->y != p3->y)
	{
		DeltaLeft.x = SCALAR32_DIV(p2->x - p1->x, p2->y - p1->y);
		DeltaLeft.u = SCALAR32_DIV(p2->u - p1->u, p2->y - p1->y);
		DeltaLeft.v = SCALAR32_DIV(p2->v - p1->v, p2->y - p1->y);
		DeltaRight.x = SCALAR32_DIV(p3->x - p1->x, p3->y - p1->y);
		DeltaRight.u = SCALAR32_DIV(p3->u - p1->u, p3->y - p1->y);
		DeltaRight.v = SCALAR32_DIV(p3->v - p1->v, p3->y - p1->y);
		if(DeltaLeft.x > DeltaRight.x)
		{
			Cur = DeltaLeft;
			DeltaLeft = DeltaRight;
			DeltaRight = Cur;
			bMiddleLeft = FALSE;
		}
		else
			bMiddleLeft = TRUE;
		Left = *p1;
		Right = *p1;
		if(SCALAR32_TO_LONG(p1->y) < 0)
		{
			Left.x -= DeltaLeft.x * SCALAR32_TO_LONG(p1->y);
			Left.u -= DeltaLeft.u * SCALAR32_TO_LONG(p1->y);
			Left.v -= DeltaLeft.v * SCALAR32_TO_LONG(p1->y);
			Right.x -= DeltaRight.x * SCALAR32_TO_LONG(p1->y);
			Right.u -= DeltaRight.u * SCALAR32_TO_LONG(p1->y);
			Right.v -= DeltaRight.v * SCALAR32_TO_LONG(p1->y);
		}
		yt = max((SCALAR16)SCALAR32_TO_LONG(p1->y), 0);
		yb = min((SCALAR16)SCALAR32_TO_LONG(p2->y), pOut->Height);
		for(y = yt; y < yb; y++)
		{
			if(Left.x != Right.x)
			{
				Cur.x = SCALAR32_DIV(Right.u - Left.u, Right.x - Left.x);
				Cur.y = SCALAR32_DIV(Right.v - Left.v, Right.x - Left.x);
				Cur.u = Left.u;
				Cur.v = Left.v;
				if(SCALAR32_TO_LONG(Left.x) < 0)
				{
					Cur.u -= Cur.x * SCALAR32_TO_LONG(Left.x);
					Cur.v -= Cur.y * SCALAR32_TO_LONG(Left.x);
				}
				xl = max((SCALAR16)SCALAR32_TO_LONG(Left.x), 0);
				xr = min((SCALAR16)SCALAR32_TO_LONG(Right.x), pOut->Width);
				pLine = pOut->pBitmap + (y << pOut->ShiftY) + xl;
				if(pColorKey)
				{
					for(x = xl; x < xr; x++)
					{
						Color = pTexture->pBitmap[((SCALAR32_TO_LONG(Cur.v) & pTexture->MaskY) << pTexture->ShiftY) + (SCALAR32_TO_LONG(Cur.u) & pTexture->MaskX)];
						if(Color != *pColorKey)
							*pLine = Color;
						Cur.u += Cur.x;
						Cur.v += Cur.y;
						pLine++;
					}
				}
				else
				{
					for(x = xl; x < xr; x++)
					{
						*pLine = pTexture->pBitmap[((SCALAR32_TO_LONG(Cur.v) & pTexture->MaskY) << pTexture->ShiftY) + (SCALAR32_TO_LONG(Cur.u) & pTexture->MaskX)];
						Cur.u += Cur.x;
						Cur.v += Cur.y;
						pLine++;
					}
				}
			}
			Left.x += DeltaLeft.x;
			Left.u += DeltaLeft.u;
			Left.v += DeltaLeft.v;
			Right.x += DeltaRight.x;
			Right.u += DeltaRight.u;
			Right.v += DeltaRight.v;
		}
	}
	else
	{
		if(p1->x <= p2->x)
		{
			DeltaLeft.x = SCALAR32_DIV(p3->x - p1->x, p3->y - p1->y);
			DeltaLeft.u = SCALAR32_DIV(p3->u - p1->u, p3->y - p1->y);
			DeltaLeft.v = SCALAR32_DIV(p3->v - p1->v, p3->y - p1->y);
			DeltaRight.x = SCALAR32_DIV(p3->x - p2->x, p3->y - p2->y);
			DeltaRight.u = SCALAR32_DIV(p3->u - p2->u, p3->y - p2->y);
			DeltaRight.v = SCALAR32_DIV(p3->v - p2->v, p3->y - p2->y);
			Left = *p1;
			Right = *p2;
			bMiddleLeft = FALSE;
		}
		else
		{
			DeltaLeft.x = SCALAR32_DIV(p3->x - p2->x, p3->y - p2->y);
			DeltaLeft.u = SCALAR32_DIV(p3->u - p2->u, p3->y - p2->y);
			DeltaLeft.v = SCALAR32_DIV(p3->v - p2->v, p3->y - p2->y);
			DeltaRight.x = SCALAR32_DIV(p3->x - p1->x, p3->y - p1->y);
			DeltaRight.u = SCALAR32_DIV(p3->u - p1->u, p3->y - p1->y);
			DeltaRight.v = SCALAR32_DIV(p3->v - p1->v, p3->y - p1->y);
			Left = *p2;
			Right = *p1;
			bMiddleLeft = TRUE;
		}
	}
	if(p1->y != p3->y && p2->y != p3->y)
	{
		if(bMiddleLeft)
		{
			DeltaLeft.x = SCALAR32_DIV(p3->x - p2->x, p3->y - p2->y);
			DeltaLeft.u = SCALAR32_DIV(p3->u - p2->u, p3->y - p2->y);
			DeltaLeft.v = SCALAR32_DIV(p3->v - p2->v, p3->y - p2->y);
			Left = *p2;
		}
		else
		{
			DeltaRight.x = SCALAR32_DIV(p3->x - p2->x, p3->y - p2->y);
			DeltaRight.u = SCALAR32_DIV(p3->u - p2->u, p3->y - p2->y);
			DeltaRight.v = SCALAR32_DIV(p3->v - p2->v, p3->y - p2->y);
			Right = *p2;
		}
		if(SCALAR32_TO_LONG(p2->y) < 0)
		{
			Left.x -= DeltaLeft.x * SCALAR32_TO_LONG(p2->y);
			Left.u -= DeltaLeft.u * SCALAR32_TO_LONG(p2->y);
			Left.v -= DeltaLeft.v * SCALAR32_TO_LONG(p2->y);
			Right.x -= DeltaRight.x * SCALAR32_TO_LONG(p2->y);
			Right.u -= DeltaRight.u * SCALAR32_TO_LONG(p2->y);
			Right.v -= DeltaRight.v * SCALAR32_TO_LONG(p2->y);
		}
		yt = max((SCALAR16)SCALAR32_TO_LONG(p2->y), 0);
		yb = min((SCALAR16)SCALAR32_TO_LONG(p3->y), pOut->Height);
		for(y = yt; y < yb; y++)
		{
			if(Left.x != Right.x)
			{
				Cur.x = SCALAR32_DIV(Right.u - Left.u, Right.x - Left.x);
				Cur.y = SCALAR32_DIV(Right.v - Left.v, Right.x - Left.x);
				Cur.u = Left.u;
				Cur.v = Left.v;
				if(SCALAR32_TO_LONG(Left.x) < 0)
				{
					Cur.u -= Cur.x * SCALAR32_TO_LONG(Left.x);
					Cur.v -= Cur.y * SCALAR32_TO_LONG(Left.x);
				}
				xl = max((SCALAR16)SCALAR32_TO_LONG(Left.x), 0);
				xr = min((SCALAR16)SCALAR32_TO_LONG(Right.x), pOut->Width);
				pLine = pOut->pBitmap + (y << pOut->ShiftY) + xl;
				if(pColorKey)
				{
					for(x = xl; x < xr; x++)
					{
						Color = pTexture->pBitmap[((SCALAR32_TO_LONG(Cur.v) & pTexture->MaskY) << pTexture->ShiftY) + (SCALAR32_TO_LONG(Cur.u) & pTexture->MaskX)];
						if(Color != *pColorKey)
							*pLine = Color;
						Cur.u += Cur.x;
						Cur.v += Cur.y;
						pLine++;
					}
				}
				else
				{
					for(x = xl; x < xr; x++)
					{
						*pLine = pTexture->pBitmap[((SCALAR32_TO_LONG(Cur.v) & pTexture->MaskY) << pTexture->ShiftY) + (SCALAR32_TO_LONG(Cur.u) & pTexture->MaskX)];
						Cur.u += Cur.x;
						Cur.v += Cur.y;
						pLine++;
					}
				}
			}
			Left.x += DeltaLeft.x;
			Left.u += DeltaLeft.u;
			Left.v += DeltaLeft.v;
			Right.x += DeltaRight.x;
			Right.u += DeltaRight.u;
			Right.v += DeltaRight.v;
		}
	}
}

void DrawVertexXYZWUV(BITMAP565* pOut, BITMAPZ16* pZBuffer, BITMAP565* pTexture, VERTEX_XYZWUV* p1, VERTEX_XYZWUV* p2, VERTEX_XYZWUV* p3, PIXEL565* pColorKey)
{
	VERTEX_XYZWUV DeltaLeft;
	VERTEX_XYZWUV DeltaRight;
	SCALAR16 x;
	SCALAR16 y;
	VERTEX_XYZWUV Left;
	VERTEX_XYZWUV Right;
	VERTEX_XYZWUV Cur;
	SCALAR16 xl;
	SCALAR16 xr;
	SCALAR16 yt;
	SCALAR16 yb;
	PIXEL565* pLine;
	PIXELZ16* pLineZ;
	PIXEL565 Color;
	BOOL bMiddleLeft;
	if(p1->y != p2->y && p1->y != p3->y)
	{
		DeltaLeft.x = SCALAR32_DIV(p2->x - p1->x, p2->y - p1->y);
		DeltaLeft.z = SCALAR32_DIV(p2->z - p1->z, p2->y - p1->y);
		DeltaLeft.w = SCALAR32_DIV(p2->w - p1->w, p2->y - p1->y);
		DeltaLeft.u = SCALAR32_DIV(p2->u - p1->u, p2->y - p1->y);
		DeltaLeft.v = SCALAR32_DIV(p2->v - p1->v, p2->y - p1->y);
		DeltaRight.x = SCALAR32_DIV(p3->x - p1->x, p3->y - p1->y);
		DeltaRight.z = SCALAR32_DIV(p3->z - p1->z, p3->y - p1->y);
		DeltaRight.w = SCALAR32_DIV(p3->w - p1->w, p3->y - p1->y);
		DeltaRight.u = SCALAR32_DIV(p3->u - p1->u, p3->y - p1->y);
		DeltaRight.v = SCALAR32_DIV(p3->v - p1->v, p3->y - p1->y);
		if(DeltaLeft.x > DeltaRight.x)
		{
			Cur = DeltaLeft;
			DeltaLeft = DeltaRight;
			DeltaRight = Cur;
			bMiddleLeft = FALSE;
		}
		else
			bMiddleLeft = TRUE;
		Left = *p1;
		Right = *p1;
		if(SCALAR32_TO_LONG(p1->y) < 0)
		{
			Left.x -= DeltaLeft.x * SCALAR32_TO_LONG(p1->y);
			Left.z -= DeltaLeft.z * SCALAR32_TO_LONG(p1->y);
			Left.w -= DeltaLeft.w * SCALAR32_TO_LONG(p1->y);
			Left.u -= DeltaLeft.u * SCALAR32_TO_LONG(p1->y);
			Left.v -= DeltaLeft.v * SCALAR32_TO_LONG(p1->y);
			Right.x -= DeltaRight.x * SCALAR32_TO_LONG(p1->y);
			Right.z -= DeltaRight.z * SCALAR32_TO_LONG(p1->y);
			Right.w -= DeltaRight.w * SCALAR32_TO_LONG(p1->y);
			Right.u -= DeltaRight.u * SCALAR32_TO_LONG(p1->y);
			Right.v -= DeltaRight.v * SCALAR32_TO_LONG(p1->y);
		}
		yt = max((SCALAR16)SCALAR32_TO_LONG(p1->y), 0);
		yb = min((SCALAR16)SCALAR32_TO_LONG(p2->y), pOut->Height);
		for(y = yt; y < yb; y++)
		{
			if(Left.x != Right.x)
			{
				Cur.x = SCALAR32_DIV(Right.u - Left.u, Right.x - Left.x);
				Cur.y = SCALAR32_DIV(Right.v - Left.v, Right.x - Left.x);
				Cur.z = SCALAR32_DIV(Right.z - Left.z, Right.x - Left.x);
				Cur.w = Left.z;
				Cur.u = Left.u;
				Cur.v = Left.v;
				if(SCALAR32_TO_LONG(Left.x) < 0)
				{
					Cur.w -= Cur.z * SCALAR32_TO_LONG(Left.x);
					Cur.u -= Cur.x * SCALAR32_TO_LONG(Left.x);
					Cur.v -= Cur.y * SCALAR32_TO_LONG(Left.x);
				}
				xl = max((SCALAR16)SCALAR32_TO_LONG(Left.x), 0);
				xr = min((SCALAR16)SCALAR32_TO_LONG(Right.x), pOut->Width);
				pLine = pOut->pBitmap + (y << pOut->ShiftY) + xl;
				if(pZBuffer)
				{
					if(pColorKey)
					{
						pLineZ = pZBuffer->pBitmap + (y << pZBuffer->ShiftY) + xl;
						for(x = xl; x < xr; x++)
						{
							if(Cur.w >= 0 && Cur.w < (SCALAR32)*pLineZ)
							{
								Color = pTexture->pBitmap[((SCALAR32_TO_LONG(Cur.v) & pTexture->MaskY) << pTexture->ShiftY) + (SCALAR32_TO_LONG(Cur.u) & pTexture->MaskX)];
								if(Color != *pColorKey)
								{
									*pLine = Color;
									*pLineZ = (PIXELZ16)Cur.w;
								}
							}
							Cur.w += Cur.z;
							Cur.u += Cur.x;
							Cur.v += Cur.y;
							pLine++;
							pLineZ++;
						}
					}
					else
					{
						pLineZ = pZBuffer->pBitmap + (y << pZBuffer->ShiftY) + xl;
						for(x = xl; x < xr; x++)
						{
							if(Cur.w >= 0 && Cur.w < (SCALAR32)*pLineZ)
							{
								*pLine = pTexture->pBitmap[((SCALAR32_TO_LONG(Cur.v) & pTexture->MaskY) << pTexture->ShiftY) + (SCALAR32_TO_LONG(Cur.u) & pTexture->MaskX)];
								*pLineZ = (PIXELZ16)Cur.w;
							}
							Cur.w += Cur.z;
							Cur.u += Cur.x;
							Cur.v += Cur.y;
							pLine++;
							pLineZ++;
						}
					}
				}
				else
				{
					if(pColorKey)
					{
						for(x = xl; x < xr; x++)
						{
							Color = pTexture->pBitmap[((SCALAR32_TO_LONG(Cur.v) & pTexture->MaskY) << pTexture->ShiftY) + (SCALAR32_TO_LONG(Cur.u) & pTexture->MaskX)];
							if(Color != *pColorKey)
								*pLine = Color;
							Cur.w += Cur.z;
							Cur.u += Cur.x;
							Cur.v += Cur.y;
							pLine++;
						}
					}
					else
					{
						for(x = xl; x < xr; x++)
						{
							*pLine = pTexture->pBitmap[((SCALAR32_TO_LONG(Cur.v) & pTexture->MaskY) << pTexture->ShiftY) + (SCALAR32_TO_LONG(Cur.u) & pTexture->MaskX)];
							Cur.w += Cur.z;
							Cur.u += Cur.x;
							Cur.v += Cur.y;
							pLine++;
						}
					}
				}
			}
			Left.x += DeltaLeft.x;
			Left.z += DeltaLeft.z;
			Left.w += DeltaLeft.w;
			Left.u += DeltaLeft.u;
			Left.v += DeltaLeft.v;
			Right.x += DeltaRight.x;
			Right.z += DeltaRight.z;
			Right.w += DeltaRight.w;
			Right.u += DeltaRight.u;
			Right.v += DeltaRight.v;
		}
	}
	else
	{
		if(p1->x <= p2->x)
		{
			DeltaLeft.x = SCALAR32_DIV(p3->x - p1->x, p3->y - p1->y);
			DeltaLeft.z = SCALAR32_DIV(p3->z - p1->z, p3->y - p1->y);
			DeltaLeft.w = SCALAR32_DIV(p3->w - p1->w, p3->y - p1->y);
			DeltaLeft.u = SCALAR32_DIV(p3->u - p1->u, p3->y - p1->y);
			DeltaLeft.v = SCALAR32_DIV(p3->v - p1->v, p3->y - p1->y);
			DeltaRight.x = SCALAR32_DIV(p3->x - p2->x, p3->y - p2->y);
			DeltaRight.z = SCALAR32_DIV(p3->z - p2->z, p3->y - p2->y);
			DeltaRight.w = SCALAR32_DIV(p3->w - p2->w, p3->y - p2->y);
			DeltaRight.u = SCALAR32_DIV(p3->u - p2->u, p3->y - p2->y);
			DeltaRight.v = SCALAR32_DIV(p3->v - p2->v, p3->y - p2->y);
			Left = *p1;
			Right = *p2;
			bMiddleLeft = FALSE;
		}
		else
		{
			DeltaLeft.x = SCALAR32_DIV(p3->x - p2->x, p3->y - p2->y);
			DeltaLeft.z = SCALAR32_DIV(p3->z - p2->z, p3->y - p2->y);
			DeltaLeft.w = SCALAR32_DIV(p3->w - p2->w, p3->y - p2->y);
			DeltaLeft.u = SCALAR32_DIV(p3->u - p2->u, p3->y - p2->y);
			DeltaLeft.v = SCALAR32_DIV(p3->v - p2->v, p3->y - p2->y);
			DeltaRight.x = SCALAR32_DIV(p3->x - p1->x, p3->y - p1->y);
			DeltaRight.z = SCALAR32_DIV(p3->z - p1->z, p3->y - p1->y);
			DeltaRight.w = SCALAR32_DIV(p3->w - p1->w, p3->y - p1->y);
			DeltaRight.u = SCALAR32_DIV(p3->u - p1->u, p3->y - p1->y);
			DeltaRight.v = SCALAR32_DIV(p3->v - p1->v, p3->y - p1->y);
			Left = *p2;
			Right = *p1;
			bMiddleLeft = TRUE;
		}
	}
	if(p1->y != p3->y && p2->y != p3->y)
	{
		if(bMiddleLeft)
		{
			DeltaLeft.x = SCALAR32_DIV(p3->x - p2->x, p3->y - p2->y);
			DeltaLeft.z = SCALAR32_DIV(p3->z - p2->z, p3->y - p2->y);
			DeltaLeft.w = SCALAR32_DIV(p3->w - p2->w, p3->y - p2->y);
			DeltaLeft.u = SCALAR32_DIV(p3->u - p2->u, p3->y - p2->y);
			DeltaLeft.v = SCALAR32_DIV(p3->v - p2->v, p3->y - p2->y);
			Left = *p2;
		}
		else
		{
			DeltaRight.x = SCALAR32_DIV(p3->x - p2->x, p3->y - p2->y);
			DeltaRight.z = SCALAR32_DIV(p3->z - p2->z, p3->y - p2->y);
			DeltaRight.w = SCALAR32_DIV(p3->w - p2->w, p3->y - p2->y);
			DeltaRight.u = SCALAR32_DIV(p3->u - p2->u, p3->y - p2->y);
			DeltaRight.v = SCALAR32_DIV(p3->v - p2->v, p3->y - p2->y);
			Right = *p2;
		}
		if(SCALAR32_TO_LONG(p2->y) < 0)
		{
			Left.x -= DeltaLeft.x * SCALAR32_TO_LONG(p2->y);
			Left.z -= DeltaLeft.z * SCALAR32_TO_LONG(p2->y);
			Left.w -= DeltaLeft.w * SCALAR32_TO_LONG(p2->y);
			Left.u -= DeltaLeft.u * SCALAR32_TO_LONG(p2->y);
			Left.v -= DeltaLeft.v * SCALAR32_TO_LONG(p2->y);
			Right.x -= DeltaRight.x * SCALAR32_TO_LONG(p2->y);
			Right.z -= DeltaRight.z * SCALAR32_TO_LONG(p2->y);
			Right.w -= DeltaRight.w * SCALAR32_TO_LONG(p2->y);
			Right.u -= DeltaRight.u * SCALAR32_TO_LONG(p2->y);
			Right.v -= DeltaRight.v * SCALAR32_TO_LONG(p2->y);
		}
		yt = max((SCALAR16)SCALAR32_TO_LONG(p2->y), 0);
		yb = min((SCALAR16)SCALAR32_TO_LONG(p3->y), pOut->Height);
		for(y = yt; y < yb; y++)
		{
			if(Left.x != Right.x)
			{
				Cur.x = SCALAR32_DIV(Right.u - Left.u, Right.x - Left.x);
				Cur.y = SCALAR32_DIV(Right.v - Left.v, Right.x - Left.x);
				Cur.z = SCALAR32_DIV(Right.z - Left.z, Right.x - Left.x);
				Cur.w = Left.z;
				Cur.u = Left.u;
				Cur.v = Left.v;
				if(SCALAR32_TO_LONG(Left.x) < 0)
				{
					Cur.w -= Cur.z * SCALAR32_TO_LONG(Left.x);
					Cur.u -= Cur.x * SCALAR32_TO_LONG(Left.x);
					Cur.v -= Cur.y * SCALAR32_TO_LONG(Left.x);
				}
				xl = max((SCALAR16)SCALAR32_TO_LONG(Left.x), 0);
				xr = min((SCALAR16)SCALAR32_TO_LONG(Right.x), pOut->Width);
				pLine = pOut->pBitmap + (y << pOut->ShiftY) + xl;
				if(pZBuffer)
				{
					if(pColorKey)
					{
						pLineZ = pZBuffer->pBitmap + (y << pZBuffer->ShiftY) + xl;
						for(x = xl; x < xr; x++)
						{
							if(Cur.w >= 0 && Cur.w < (SCALAR32)*pLineZ)
							{
								Color = pTexture->pBitmap[((SCALAR32_TO_LONG(Cur.v) & pTexture->MaskY) << pTexture->ShiftY) + (SCALAR32_TO_LONG(Cur.u) & pTexture->MaskX)];
								if(Color != *pColorKey)
								{
									*pLine = Color;
									*pLineZ = (PIXELZ16)Cur.w;
								}
							}
							Cur.w += Cur.z;
							Cur.u += Cur.x;
							Cur.v += Cur.y;
							pLine++;
							pLineZ++;
						}
					}
					else
					{
						pLineZ = pZBuffer->pBitmap + (y << pZBuffer->ShiftY) + xl;
						for(x = xl; x < xr; x++)
						{
							if(Cur.w >= 0 && Cur.w < (SCALAR32)*pLineZ)
							{
								*pLine = pTexture->pBitmap[((SCALAR32_TO_LONG(Cur.v) & pTexture->MaskY) << pTexture->ShiftY) + (SCALAR32_TO_LONG(Cur.u) & pTexture->MaskX)];
								*pLineZ = (PIXELZ16)Cur.w;
							}
							Cur.w += Cur.z;
							Cur.u += Cur.x;
							Cur.v += Cur.y;
							pLine++;
							pLineZ++;
						}
					}
				}
				else
				{
					if(pColorKey)
					{
						for(x = xl; x < xr; x++)
						{
							Color = pTexture->pBitmap[((SCALAR32_TO_LONG(Cur.v) & pTexture->MaskY) << pTexture->ShiftY) + (SCALAR32_TO_LONG(Cur.u) & pTexture->MaskX)];
							if(Color != *pColorKey)
								*pLine = Color;
							Cur.w += Cur.z;
							Cur.u += Cur.x;
							Cur.v += Cur.y;
							pLine++;
						}
					}
					else
					{
						for(x = xl; x < xr; x++)
						{
							*pLine = pTexture->pBitmap[((SCALAR32_TO_LONG(Cur.v) & pTexture->MaskY) << pTexture->ShiftY) + (SCALAR32_TO_LONG(Cur.u) & pTexture->MaskX)];
							Cur.w += Cur.z;
							Cur.u += Cur.x;
							Cur.v += Cur.y;
							pLine++;
						}
					}
				}
			}
			Left.x += DeltaLeft.x;
			Left.z += DeltaLeft.z;
			Left.w += DeltaLeft.w;
			Left.u += DeltaLeft.u;
			Left.v += DeltaLeft.v;
			Right.x += DeltaRight.x;
			Right.z += DeltaRight.z;
			Right.w += DeltaRight.w;
			Right.u += DeltaRight.u;
			Right.v += DeltaRight.v;
		}
	}
}

void DrawVertexArrayXY(BITMAP565* pOut, VERTEX_XY* pData, DWORD Count, DWORD CullMode, PIXEL565 Color)
{
	SCALAR32 Direction;
	VERTEX_XY* pEnd;
	VERTEX_XY* pTemp[3];
	switch(CullMode)
	{
	case VERTEX_CULL_NONE:
		Direction = 0;
		break;
	case VERTEX_CULL_LEFT:
		Direction = -1;
		break;
	case VERTEX_CULL_RIGHT:
		Direction = 1;
		break;
	}
	pEnd = pData + Count - 2;
	while(pData < pEnd)
	{
		if((SCALAR32_MUL(pData[1].x - pData[0].x, pData[2].y - pData[1].y) - SCALAR32_MUL(pData[2].x - pData[1].x, pData[1].y - pData[0].y)) * Direction >= 0)
		{
			if(pData[0].y <= pData[1].y && pData[0].y <= pData[2].y)
			{
				pTemp[0] = &pData[0];
				if(pData[1].y <= pData[2].y)
				{
					pTemp[1] = &pData[1];
					pTemp[2] = &pData[2];
				}
				else
				{
					pTemp[1] = &pData[2];
					pTemp[2] = &pData[1];
				}
			}
			else if(pData[1].y <= pData[0].y && pData[1].y <= pData[2].y)
			{
				pTemp[0] = &pData[1];
				if(pData[0].y <= pData[2].y)
				{
					pTemp[1] = &pData[0];
					pTemp[2] = &pData[2];
				}
				else
				{
					pTemp[1] = &pData[2];
					pTemp[2] = &pData[0];
				}
			}
			else
			{
				pTemp[0] = &pData[2];
				if(pData[0].y <= pData[1].y)
				{
					pTemp[1] = &pData[0];
					pTemp[2] = &pData[1];
				}
				else
				{
					pTemp[1] = &pData[1];
					pTemp[2] = &pData[0];
				}
			}
			if(pData[0].y != pData[1].y || pData[1].y != pData[2].y)
				DrawVertexXY(pOut, pTemp[0], pTemp[1], pTemp[2], Color);
		}
		Direction *= -1;
		pData++;
	}
}

void DrawVertexArrayXYUV(BITMAP565* pOut, BITMAP565* pTexture, VERTEX_XYUV* pData, DWORD Count, DWORD CullMode, PIXEL565* pColorKey)
{
	SCALAR32 Direction;
	VERTEX_XYUV* pEnd;
	VERTEX_XYUV* pTemp[3];
	switch(CullMode)
	{
	case VERTEX_CULL_NONE:
		Direction = 0;
		break;
	case VERTEX_CULL_LEFT:
		Direction = -1;
		break;
	case VERTEX_CULL_RIGHT:
		Direction = 1;
		break;
	}
	pEnd = pData + Count - 2;
	while(pData < pEnd)
	{
		if((SCALAR32_MUL(pData[1].x - pData[0].x, pData[2].y - pData[1].y) - SCALAR32_MUL(pData[2].x - pData[1].x, pData[1].y - pData[0].y)) * Direction >= 0)
		{
			if(pData[0].y <= pData[1].y && pData[0].y <= pData[2].y)
			{
				pTemp[0] = &pData[0];
				if(pData[1].y <= pData[2].y)
				{
					pTemp[1] = &pData[1];
					pTemp[2] = &pData[2];
				}
				else
				{
					pTemp[1] = &pData[2];
					pTemp[2] = &pData[1];
				}
			}
			else if(pData[1].y <= pData[0].y && pData[1].y <= pData[2].y)
			{
				pTemp[0] = &pData[1];
				if(pData[0].y <= pData[2].y)
				{
					pTemp[1] = &pData[0];
					pTemp[2] = &pData[2];
				}
				else
				{
					pTemp[1] = &pData[2];
					pTemp[2] = &pData[0];
				}
			}
			else
			{
				pTemp[0] = &pData[2];
				if(pData[0].y <= pData[1].y)
				{
					pTemp[1] = &pData[0];
					pTemp[2] = &pData[1];
				}
				else
				{
					pTemp[1] = &pData[1];
					pTemp[2] = &pData[0];
				}
			}
			if(pData[0].y != pData[1].y || pData[1].y != pData[2].y)
				DrawVertexXYUV(pOut, pTexture, pTemp[0], pTemp[1], pTemp[2], pColorKey);
		}
		Direction *= -1;
		pData++;
	}
}

void DrawVertexArrayXYZWUV(BITMAP565* pOut, BITMAPZ16* pZBuffer, BITMAP565* pTexture, VERTEX_XYZWUV* pData, DWORD Count, DWORD CullMode, PIXEL565* pColorKey)
{
	SCALAR32 Direction;
	VERTEX_XYZWUV* pEnd;
	VERTEX_XYZWUV* pTemp[3];
	switch(CullMode)
	{
	case VERTEX_CULL_NONE:
		Direction = 0;
		break;
	case VERTEX_CULL_LEFT:
		Direction = -1;
		break;
	case VERTEX_CULL_RIGHT:
		Direction = 1;
		break;
	}
	pEnd = pData + Count - 2;
	while(pData < pEnd)
	{
		if((SCALAR32_MUL(pData[1].x - pData[0].x, pData[2].y - pData[1].y) - SCALAR32_MUL(pData[2].x - pData[1].x, pData[1].y - pData[0].y)) * Direction >= 0)
		{
			if(pData[0].y <= pData[1].y && pData[0].y <= pData[2].y)
			{
				pTemp[0] = &pData[0];
				if(pData[1].y <= pData[2].y)
				{
					pTemp[1] = &pData[1];
					pTemp[2] = &pData[2];
				}
				else
				{
					pTemp[1] = &pData[2];
					pTemp[2] = &pData[1];
				}
			}
			else if(pData[1].y <= pData[0].y && pData[1].y <= pData[2].y)
			{
				pTemp[0] = &pData[1];
				if(pData[0].y <= pData[2].y)
				{
					pTemp[1] = &pData[0];
					pTemp[2] = &pData[2];
				}
				else
				{
					pTemp[1] = &pData[2];
					pTemp[2] = &pData[0];
				}
			}
			else
			{
				pTemp[0] = &pData[2];
				if(pData[0].y <= pData[1].y)
				{
					pTemp[1] = &pData[0];
					pTemp[2] = &pData[1];
				}
				else
				{
					pTemp[1] = &pData[1];
					pTemp[2] = &pData[0];
				}
			}
			if(pData[0].y != pData[1].y || pData[1].y != pData[2].y)
				DrawVertexXYZWUV(pOut, pZBuffer, pTexture, pTemp[0], pTemp[1], pTemp[2], pColorKey);
		}
		Direction *= -1;
		pData++;
	}
}

void DrawVertexIndexArrayXYZWUV(BITMAP565* pOut, BITMAPZ16* pZBuffer, BITMAP565* pTexture, VERTEX_XYZWUV* pData, DWORD* pIndex, DWORD Count, DWORD CullMode, PIXEL565* pColorKey)
{
	SCALAR32 Direction;
	DWORD* pEnd;
	VERTEX_XYZWUV* pVertex[3];
	VERTEX_XYZWUV* pTemp[3];
	switch(CullMode)
	{
	case VERTEX_CULL_NONE:
		Direction = 0;
		break;
	case VERTEX_CULL_LEFT:
		Direction = -1;
		break;
	case VERTEX_CULL_RIGHT:
		Direction = 1;
		break;
	}
	pEnd = pIndex + Count * 3;
	while(pIndex < pEnd)
	{
		pVertex[0] = &pData[pIndex[0]];
		pVertex[1] = &pData[pIndex[1]];
		pVertex[2] = &pData[pIndex[2]];
		if((SCALAR32_MUL(pVertex[1]->x - pVertex[0]->x, pVertex[2]->y - pVertex[1]->y) - SCALAR32_MUL(pVertex[2]->x - pVertex[1]->x, pVertex[1]->y - pVertex[0]->y)) * Direction >= 0)
		{
			if(pVertex[0]->y <= pVertex[1]->y && pVertex[0]->y <= pVertex[2]->y)
			{
				pTemp[0] = pVertex[0];
				if(pVertex[1]->y <= pVertex[2]->y)
				{
					pTemp[1] = pVertex[1];
					pTemp[2] = pVertex[2];
				}
				else
				{
					pTemp[1] = pVertex[2];
					pTemp[2] = pVertex[1];
				}
			}
			else if(pVertex[1]->y <= pVertex[0]->y && pVertex[1]->y <= pVertex[2]->y)
			{
				pTemp[0] = pVertex[1];
				if(pVertex[0]->y <= pVertex[2]->y)
				{
					pTemp[1] = pVertex[0];
					pTemp[2] = pVertex[2];
				}
				else
				{
					pTemp[1] = pVertex[2];
					pTemp[2] = pVertex[0];
				}
			}
			else
			{
				pTemp[0] = pVertex[2];
				if(pVertex[0]->y <= pVertex[1]->y)
				{
					pTemp[1] = pVertex[0];
					pTemp[2] = pVertex[1];
				}
				else
				{
					pTemp[1] = pVertex[1];
					pTemp[2] = pVertex[0];
				}
			}
			if(pVertex[0]->y != pVertex[1]->y || pVertex[1]->y != pVertex[2]->y)
				DrawVertexXYZWUV(pOut, pZBuffer, pTexture, pTemp[0], pTemp[1], pTemp[2], pColorKey);
		}
		pIndex += 3;
	}
}

